/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault
 *
 * de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.CollectDeBlocksDefault wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion;

/**
 * Legt die Liste der FG/ID Kombinationen fest, für die DE-Blöcke nicht direkt an die Streckenstation versandt werden,
 * sondern erst nach Triggerung. Damit lassen sich z. B. die DE der FG 4 (Anzeigen) für einen Anzeigequerschnitt
 * (physikalisch die DE der FG 4 an einem EAK) in einem Einzeltelegramm mit mehreren DE-Blöcken für diesen AQ versenden.
 * Die Triggerung erfolgt entweder für ein Steuermodul (erzeugt dann jeweils Trigger für die angeschlossenen EAK) oder
 * ein einzelnes EAK.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */

public class CollectDeBlocksDefault {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/**
	 * Enthält je Eintrag jeweils Funktionsgruppe, Typ und ID des De-Blocks, welches nicht sofort sondern erst nach
	 * Triggerung am EAK bzw. SM in einem Einzeltelegramm mit mehreren DE-Blöcken versandt werden soll.
	 */
	private static final int[][] collectDeBlocks = new int[][]{
			//  FG, Typ, ID
			{4, 32, 3}, // WVZ-Grundeinstellung (bis TLS 2002 )
			{4, 33, 3}, // Grundeinstellung (ab TLS 2002 )
			{4, 42, 3}, // Grundprogramm (bis TLS 2002 )
			{4, 43, 3}, // Codeliste (ab TLS 2002 )
			{4, 44, 3}, // Codedefinition (ab TLS 2002 )
			{4, 48, 5}, // WVZ-Stellzustand (bis TLS 2002 )
			{4, 50, 5}, // Wechseltext (bis TLS 2002 )
			{4, 55, 5}, // Stellzustand (ab TLS 2002 )
			{4, 58, 5}, // Stellprogramm (bis TLS 2002 )
	};

// ------------------------------------------------------------------------
// STATISCHE METHODEN
// ------------------------------------------------------------------------

	public static int[][] getCollectDeBlocks() {
		return collectDeBlocks;
	}
}
