/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.DeConversionModul
 *
 * de.bsvrz.kex.tls.osi7.conversion.DeConversionModul is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.DeConversionModul is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.DeConversionModul.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.DeConversionModul.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.DeConversionModul ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.DeConversionModul wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.osi7.conversion;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.tls.TlsApplication;
import de.bsvrz.kex.tls.osi7.CollectDeBlocks;
import de.bsvrz.kex.tls.osi7.De;
import de.bsvrz.kex.tls.osi7.DeBlock;
import de.bsvrz.kex.tls.osi7.DeBlockToSend;
import de.bsvrz.kex.tls.osi7.Eak;
import de.bsvrz.kex.tls.osi7.MessageSender;
import de.bsvrz.kex.tls.osi7.TlsApplicationLayer;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

//~ JDK IMPORTE ===============================================================

//~ KLASSEN ===================================================================

/**
 * Abstrakte Umsetzungsklasse, die das {@link DeConversionInterface} und zusätzlich ein Reihe von notwendigen Methoden
 * zentral implementiert. Von dieser abstrakten Klasse werden alle konkreten Umsetzungsklassen direkt oder indirekt
 * abgeleitet.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class DeConversionModul implements DeConversionInterface, ClientReceiverInterface, ClientSenderInterface {

    /** Index für den Zugriff auf den Aspekt in den Objekten davToTls und tlsToDav */
    public static final int ASP = 1;

    // ------------------------------------------------------------------------
    // VARIABLENDEKLARATIONEN
    // ------------------------------------------------------------------------

    /** Index für den Zugriff auf die Attributgruppe in den Objekten davToTls und tlsToDav */
    public static final int ATG = 0;

    /** Index für den Zugriff auf die Funktionsgruppe in den Objekten davToTls und tlsToDav */
    public static final int FG = 2;

    /** Index für den Zugriff auf die ID in den Objekten davToTls und tlsToDav */
    public static final int ID = 4;

    /** Index für den Zugriff auf den ObjektTyp in den Objekten davToTls und tlsToDav */
    public static final int OT = 5;

    // Sonstige ---------------------------------------------------------------

    /** Index für den Zugriff auf den Typ in den Objekten davToTls und tlsToDav */
    public static final int TYP = 3;

    /** DebugLogger für Debug-Ausgaben */
    private static final Debug debug = Debug.getLogger();

    /**
     * Formatstring für die Formatierung der Zahlenausgaben für die Fg, Typ und Id Nummernkonvertierung beim ermitteln
     * des Methodennamen
     */
    private static final DecimalFormat _numberFormat = new DecimalFormat("000");

    //~ FELDER ================================================================

    /**
     * Hashtable mit dem jeweils letzten Stand eines versandten De-Blocks für jede Knotennummer-DeNummer-Fg-Typ-Id
     * Kombination. Uber die Methode {@link #sendInitialDeBlock} kann dann z.B. die aktuelle Einstellung für die
     * Kanalsteuerung nach einer De-Gutmeldung verschickt werden. Die Methode greift dabei über den beim speichern
     * verwendeten Hashkey des DeBlocks {@link DeBlock#getLongKey} auf den hier gespeicherten DeBlock zu. Alle
     * versandten De-Blöcke werden nach dem Versand (tatsächlich wenn sie in der Queue zum versenden stehen, ob sie
     * wirklich versandt wurden ist nicht relevant) mit dem {@link DeBlock#getLongKey} in dieser Tabelle gespeichert.
     */
    Map<String, DeBlock> _lastSendDeBlock = new Hashtable<String, DeBlock>();

    /**
     * Liste, in der das Verhalten beim Empfang von Attributgruppen festgelegt wird. Die Liste wird zur Laufzeit durch
     * Aufrufe der Methode {@link #subscribeAllData} initialisiert, die wiederum von konkreten Instanzen dieser Klasse
     * aufgerufen werden. Die Listeneinträge sind Object[] mit jeweils folgendem Aufbau:
     * <pre>
     * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
     * z.B.
     * "atg.tlsGloKanalSteuerung", "asp.parameterSoll", "1", "29", "2",
     * "typ.deLve"
     * </pre>
     * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
     * Empfangsanmeldung als Empfänger.</li> <li> In der Methode {@link #update} wird für jede zur
     * Attributgruppen/Aspket-Kombination passende Kombination eine Methode aufgerufen, deren Namen aus dem
     * Tabelleneintrag wie folgt gebildet wird (hier für den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspParameterSoll(ResultData)</code><br>
     * Dabei enthält <code>ResultData</code> die Daten der zu verarbeitenden ATG/ASP-Kombination.</li> <li> Für jede
     * Attributguppen/Aspekt Kombination erfolgt eine Empfangsanmeldung als Empfänger. Sind mehrere passende Eintrage
     * vorhanden, werden entsprechend mehrere Methodenaufrufe durchgeführt. Dadurch können mehrere ATG/ASP-Kombinationen
     * in einen De-Block abgebildet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eines
     * Fachmodells):</li> <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
     */
    private List<Object[ ]> _davToTlsList = new ArrayList<Object[ ]>();

    /**
     * Liste, in der das Verhalten beim Empfang von TLS-Daten festgelegt wird. Die Liste wird zur Laufzeit durch Aufrufe
     * der Methode {@link #subscribeAllData} initialisiert, die wiederum von konkreten Instanzen dieser Klasse
     * aufgerufen werden. Die Listeneinträge sind Object[] mit jeweils folgendem Aufbau:
     * <pre>
     * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
     * z.B.
     * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
     * </pre>
     * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
     * Sendeanmeldung als Quelle.</li> <li> In der Methode {@link #processDeBlock} wird für jede zum De-Block passende
     * Fg/Id/Typ-Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag wie folgt gebildet wird (hier
     * für den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspTlsAntwort(DeBlock, Data,
     * DataDescription)</code><br> Dabei enthält der <code>deBlock</code> die Daten des zu verarbeitenden De-Blocks und
     * <code>data</code> ein Datenobjekt der passenden Attributgruppen/Aspekt-Kombination, welchem direkt die Inhalte
     * des De-Blocks zugewiesen werden können.<code>DataDescription</code> enthält die passende {@link DataDescription}
     * für die zu sendenden Daten.</li> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine Sendeanmeldung als
     * Quelle. Sind mehrere passende Eintrage vorhanden, werden entsprechend mehrere Methodenaufrufe durchgeführt.
     * Dadurch kann ein De-Block mehrfach ausgewert und an verschiedene Attributgruppen weitergerleitet werden (z.B.
     * protokollnahe TLS-Attributgruppe und Attributgruppe eine Fachmodells):</li> <li> Ein Eintrag von "255" entspricht
     * dabei einer "Wildcard" (passt immer). </ul>
     */
    private List<Object[ ]> _tlsToDavList = new ArrayList<Object[ ]>();

    /** Zugriff auf das Objekt zum Versenden von Betriebsmeldungen */
    MessageSender _messageSender = MessageSender.getInstance();

    /** Legt fest, ob Umsetungsmethoden ausgeführt werden sollen oder nicht */
    private boolean _isPassivierungUmsetzungsMethoden = false;

    /** Flag, um doppelte Sendeanmeldungen von FG4-DE-Sammelkanälen zu verhindern */
    private boolean _isFg4AngemeldetAlsSammelKanal = false;

    /** Flag, um doppelte Sendeanmeldungen von FG4-DE-Clustern zu verhindern */
    private boolean _isFg4AngemeldetAlsCluster = false;

    /** Legt fest, ob Betriebsmeldungen in Umsetzungsmethoden verschickt werden oder nicht */
    private boolean _isBetriebsMeldungenErzeugen = true;

    /** Flag, um doppelte Anmeldungen zur Verwaltung der Passivierung von Umsetungsmethoden zu verhindern */
    private boolean _isAngemeldetPassivierungUmsetzungsMethoden = false;

    /** Flag, um doppelte Anmeldungen zur Verwaltung der Betriebsmeldungen in Umsetzungsmethoden zu verhindern */
    private boolean _isAngemeldetBetriebsMeldungenErzeugen = false;

    // Private ----------------------------------------------------------------

    /** ToDo: Beschreibung Feld */
    private CollectDeBlocks _collectDeBlocks;

    /** Datenmodell der Konfiguration. */
    public DataModel _config;

    /** Methode im UmsetzungsModul, welche dynamisch zur Laufzeit ermittelt wird. */
    Method _conversionMethod;

    /** Argumente der Methode im UmsetzungsModul, welche dynamisch zur Laufzeit ermittelt wird. */
    Class[ ] _conversionMethodArguments;

    /** MethodenName im UmsetzungsModul, welcher dynamisch zur Laufzeit ermittelt wird. */
    String _conversionMethodName;

    /** Objektreferenz auf das aufrufende EakConversionInterface-Objekt. */

    // private final EakConversionInterface _eakConversionModul;

    /** Name des UmsetzungsModul für die Konvertierung der Telegramme in Attributgruppen (und zurück) */
    String _conversionModulName;

    /** DaV Interface. */
    ClientDavInterface _daf;

    /** Objektreferenz auf das aufrufende De-Objekt. */
    protected De _de;

    /** Objektreferenz auf das aufrufende dem De-Objekt übergeordnete Eak-Objekt. */
    protected Eak _eak;

    /**
     * Objektreferenz auf das Eak-Umwandlaungs-Objekt, welches für die Auswertung De-übergreifender Informationen
     * (Zeitstempel etc. ) zuständig ist.
     */
    protected EakConversionInterface _eakConversionModul;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    // ------------------------------------------------------------------------
    // KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
    // ------------------------------------------------------------------------

    /**
     * Erzeugt ein Objekt vom Typ DeConversionModul
     *
   	 * @param daf DaV
   	 * @param eak Eak
   	 * @param de De
   	 */
    public DeConversionModul(ClientDavInterface daf, Eak eak, De de) {
        _daf                = daf;
        _de                 = de;
        _eak                = eak;
        _eakConversionModul = _eak.getConversionModul();
        _config             = _daf.getDataModel();

        // Instanz der Klasse CollectDeBlock holen.
        _collectDeBlocks = CollectDeBlocks.getInstance();
    }

    //~ METHODEN ==============================================================

    // ------------------------------------------------------------------------
    // SONSTIGE METHODEN
    // ------------------------------------------------------------------------

    /**
     * Bildet den Methodennamen aus den Einzelangaben von davToTls und tlsToDav.
     *
     * @param tlsSendReceiveDescription Object[] entsprechend {@link #_tlsToDavList} und {@link #_davToTlsList}.
     *
     * @return Methodenname entsprechend der Beschreibung unter {@link #_tlsToDavList} und {@link #_davToTlsList}.
     */
    private String buildMethodName(Object[ ] tlsSendReceiveDescription) {
        int    fg         = Integer.parseInt((String) (tlsSendReceiveDescription[ FG ]));
        int    typ        = Integer.parseInt((String) (tlsSendReceiveDescription[ TYP ]));
        int    id         = Integer.parseInt((String) (tlsSendReceiveDescription[ ID ]));
        String methodName = "Fg" + _numberFormat.format(fg) + "Typ" + _numberFormat.format(typ) + "Id" + _numberFormat.format(id) + "Atg" + ((String) tlsSendReceiveDescription[ ATG ]).substring(4, 5).toUpperCase() + ((String) tlsSendReceiveDescription[ ATG ]).substring(5, ((String) tlsSendReceiveDescription[ ATG ]).length()) + "Asp" + ((String) tlsSendReceiveDescription[ ASP ]).substring(4, 5).toUpperCase() +
                            ((String) tlsSendReceiveDescription[ ASP ]).substring(5, ((String) tlsSendReceiveDescription[ ASP ]).length());

        return methodName;
    }

    // --------------------- Interface ClientSenderInterface ---------------------

    /**
     * Sendesteuerung des Datenverteilers an die Applikation. Diese Methode muss von der Applikation implementiert
     * werden, um den Versand von Daten zu starten bzw. anzuhalten. Der Datenverteiler signalisiert damit einer Quelle
     * oder einem Sender dass mindestens ein Abnehmer bzw. kein Abnehmer mehr für die zuvor angemeldeten Daten vorhanden
     * ist. Die Quelle wird damit aufgefordert den Versand von Daten zu starten bzw. zu stoppen.
     *
     * @param object          Das in der zugehörigen Sendeanmeldung angegebene Objekt, auf das sich die Sendesteuerung
     *                        bezieht.
     * @param dataDescription Beschreibende Informationen zu den angemeldeten Daten auf die sich die Sendesteuerung
     *                        bezieht.
     * @param state           Status der Sendesteuerung. Kann einen der Werte <code>START_SENDING</code>,
     *                        <code>STOP_SENDING</code>, <code>STOP_SENDING_NO_RIGHTS</code>,
     *                        <code>STOP_SENDING_NOT_A_VALID_SUBSCRIPTION</code> enthalten.
     *
     * @see #START_SENDING
     * @see #STOP_SENDING
     * @see #STOP_SENDING_NO_RIGHTS
     * @see #STOP_SENDING_NOT_A_VALID_SUBSCRIPTION
     */
    public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
        debug.finest("Sendesteuerung erhalten: " + Debug.NEWLINE + "Objekt : " + object + Debug.NEWLINE + "Daten  : " + dataDescription + Debug.NEWLINE + "Status : " + state);
    }

    //~ GET METHODEN ==========================================================

    // ------------------------------------------------------------------------
    // GETTER / SETTER METHODEN
    // ------------------------------------------------------------------------

    /**
     * Ermöglicht den Zugriff auf die aktuelle Datenverteilerverbindung.
     *
     * @return Gibt das aktuelle aktuelle Datenverteilerverbindung zurück.
     */
    public ClientDavInterface getClientDavInterface() {
        return _daf;
    }

    /**
     * Liefert das aktuelle Datenmodell
     *
     * @return aktuelles Datenmodell
     */
    public DataModel getConfig() {
        return _config;
    }

    /**
     * Liste, in der das Verhalten beim Senden von TLS-Daten festgelegt wird. Die Liste wird zur Laufzeit durch Aufrufe
     * der Methode {@link #subscribeAllData} initialisiert, die wiederum von konkreten Instanzen dieser Klasse
     * aufgerufen werden. Die Listeneinträge sind Object[] mit jeweils folgendem Aufbau:
     * <pre>
     * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
     * z.B.
     * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
     * </pre>
     *
     * @return Die Listeneinträge mit obigem Aufbau
     */
    public List<Object[ ]> getDavToTlsList() {
        return _davToTlsList;
    }

    /**
     * Liste, in der das Verhalten beim Empfang von TLS-Daten festgelegt wird. Die Liste wird zur Laufzeit durch Aufrufe
     * der Methode {@link #subscribeAllData} initialisiert, die wiederum von konkreten Instanzen dieser Klasse
     * aufgerufen werden. Die Listeneinträge sind Object[] mit jeweils folgendem Aufbau:
     * <pre>
     * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
     * z.B.
     * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
     * </pre>
     *
     * @return Die Listeneinträge mit obigem Aufbau
     */
    public List<Object[ ]> getTlsToDavList() {
        return _tlsToDavList;
    }

    /**
     * Ermittelt, ob ein DE-Block (für die DE255 bzw. die DE0 und die Clusterkanäle) überhaupt an die DE
     * gesendet werden darf. Dies ist z. B.
     * in der Regel für die DE 255/DE0 bei den meisten Daten nicht erlaubt. Für diese DE werden aber mit der
     * Parametrierung anfänglich Parameterdaten publiziert, die nicht versandt werden dürfen.
     *
     * @param de      Die De, für der der DE-Block geprüft wird
     * @param deBlock Der zu prüfende DE-Block
     *
     * @return <code>true</code>, wenn der DE-Block versendet werden muas (darf), sonst <code>false</code>.
     */
    private boolean isAllowedToSend(De de, DeBlock deBlock) {

    	// Abruftelegramme (haben keine Nutzdaten) dürfen immer versendet werden
        int laengeNutzdaten = deBlock.getDeData().length();

        if (laengeNutzdaten == 0) {
            return true;
        }

        // Bei Clusterkanälen den Versand von Kanalsteuerung (Typ 29) unterdrücken
        if (de.isCluster()) {
            if (deBlock.getTyp() == 29) {
                return false;
            }
        }

        if ((deBlock.getDeChannel() == 255)
            || (deBlock.getDeChannel() == 0)) {
            int fg  = deBlock.getFg();
            int typ = deBlock.getTyp();

            switch (fg) {
            case 254 :
                switch (typ) {
                case 29 :  // Kanalsteuerung
                case 32 :  // Statische Gerätekenndaten
                case 34 :  // Konfigurationstabelle
                    return false;
                }

                break;

            case 1 :
                switch (typ) {
                case 29 :  // Kanalsteuerung
                case 32 :  // LVE-Betriebsparameter
                case 33 :  // EFZ-Betriebsparameter
                case 34 :  // SVE-Betriebsparameter
                case 36 :  // Geografische Kenndaten
                    return false;
                }

                break;

            case 2 :
                switch (typ) {
                case 29 :  // Kanalsteuerung
                case 32 :  // Betriebsparameter
                case 36 :  // Geografische Kenndaten
                case 37 :  // Grenzwerte Achslasten und Gesamtgewichte
                case 38 :  // Parameter Achslast-/ Gewichtsklassen 5+1
                case 39 :  // Parameter Achslast-/ Gewichtsklassen 8+1
                    return false;
                }

                break;

            case 3 :
                switch (typ) {
                case 29 :  // Kanalsteuerung
                case 32 :  // Betriebsparameter
                case 36 :  // Geografische Kenndaten
                    return false;
                }

                break;

            case 4 :
                switch (typ) {
                case 17 :  // Betriebsart
                case 29 :  // Kanalsteuerung
                case 32 :  // WVZ-Grundeinstellung
                case 33 :  // Grundeinstellung
                case 36 :  // Geografische Kenndaten
                case 42 :  // Grundprogramm
                case 43 :  // Codelisten
                case 44 :  // Codedefinition
                case 45 :  // Bilddefinitionen
                case 48 :  // WVZ-Stellzustand
                case 49 :  // Helligkeit
                case 50 :  // Wechseltext
                case 55 :  // Stellzustand
                case 58 :  // Stellprogramm
                    return false;
                }

                break;

            case 5 :
                switch (typ) {
                case 17 :  // Betriebsart
                case 29 :  // Kanalsteuerung
                case 32 :  // Programmdaten
                case 33 :  // Programmnummern
                case 36 :  // Geografische Kenndaten
                case 48 :  // WVZ-Stellzustand
                case 49 :  // Prismen-Stellzustand
                case 50 :  // Wechseltext
                case 51 :  // Helligkeit
                    return false;
                }

                break;

            case 6 :
                switch (typ) {
                case 29 :  // Kanalsteuerung
                case 32 :  // Betriebsparameter
                case 36 :  // Geografische Kenndaten
                    return false;
                }

                break;

            case 9 :
                switch (typ) {
                case 29 :  // Kanalsteuerung
                case 32 :  // Betriebsparameter
                case 33 :  // Betriebsparameter ALINEA1
                case 34 :  // Betriebsparameter Rotfahrerzähler
                case 35 :  // Betriebsparameter ALINEA2
                case 36 :  // Geografische Kenndaten
                case 48 :  // Signalplan
                case 49 :  // Helligkeit
                    return false;
                }

                break;
            }
        }

        return true;
    }

    /**
     * ToDo: Beschreibung der Methode
     *
     *
     * @return isBetriebsMeldungenErzeugen
     */
    public boolean isBetriebsMeldungenErzeugen() {
        return _isBetriebsMeldungenErzeugen;
    }

    /**
     * ToDo: Beschreibung der Methode
     *
     *
     * @return isPassivierungUmsetzungsMethoden
     */
    public boolean isPassivierungUmsetzungsMethoden() {
        return _isPassivierungUmsetzungsMethoden;
    }

    /**
     * Diese Methode muss von der Applikation implementiert werden, um zu signalisieren, ob Sendesteuerungen erwünscht
     * sind und mit der Methode <code>dataRequest</code> verarbeitet werden.
     *
     * @param object          Das in der zugehörigen Sendeanmeldung angegebene System-Objekt.
     * @param dataDescription Die in der zugehörigen Sendeanmeldung angegebenen beschreibenden Informationen der
     *                        angemeldeten Daten.
     *
     * @return <code>true</code>, falls Sendesteuerungen gewünscht sind, sonst <code>false</code>.
     */
    public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
        return true;
    }

    //~ METHODEN ==============================================================

    // --------------------- Interface DeConversionInterface ---------------------

    /**
     * Verarbeitet einen DeBlock.
     *
     * @param deBlock Der in eine Attributgruppe umzusetzende {@link DeBlock}.
     */
    public void processDeBlock(DeBlock deBlock) {
        ResultData result;
        boolean    isConverted = false;  // wurde DeBlock konvertiert ?

        debug.finest("DeBlock empfangen für Weiterverarbeitung (TLS->DaV): " + Debug.NEWLINE + deBlock);

        try {
            Iterator<Object[ ]> iterator = _tlsToDavList.iterator();

            while (iterator.hasNext()) {
                Object[ ] tlsToDav = iterator.next();
                int       fg       = Integer.parseInt((String) (tlsToDav[ FG ]));
                int       typ      = Integer.parseInt((String) (tlsToDav[ TYP ]));
                int       id       = Integer.parseInt((String) (tlsToDav[ ID ]));

                // Passende Fg/Typ/Id Signatur suchen...
                if (((deBlock.getFg() == fg) || (fg == 255))
                    && ((deBlock.getTyp() == typ) || (typ == 255))
                    && ((deBlock.getId() == id) || (id == 255))) {
                    isConverted = true;  // es gibt mindestens einen passenden Konvertierungseintrag

                    // ... daraus den Methodennamen bilden...
                    _conversionMethodName = buildMethodName(tlsToDav);
                    _conversionModulName  = this.toString();

                    // ... und ein zur Attributgruppe passendes Data-Objekt erzeugen...
                    AttributeGroup  attributeGroup  = _config.getAttributeGroup((String) tlsToDav[ ATG ]);
                    Aspect          aspect          = _config.getAspect((String) tlsToDav[ ASP ]);
                    DataDescription dataDescription = new DataDescription(attributeGroup, aspect);
                    Data            data            = _daf.createData(attributeGroup);

                    debug.finest("Ermittelter Methodenname für Umsetzung: " + _conversionMethodName + Debug.NEWLINE + "De-Block: " + Debug.NEWLINE + deBlock.toString());

                    // ... und die Methode aufrufen mit den Parametern deBlock und data.
                    _conversionMethodArguments = new Class[ ] { DeBlock.class, Data.class, DataDescription.class };
                    _conversionMethod          = this.getClass().getMethod(_conversionMethodName, _conversionMethodArguments);

                    // REWIND !!!! für mehrfaches auswerten des DE Blocks.
                    deBlock.getDeData().rewindReadPosition();
                    result = (ResultData) _conversionMethod.invoke(this, deBlock, data, dataDescription);

                    // TBD: Unterscheidung zwischen Fehlern -> dann leeren Datensatz schicken, um zu dokumentieren,
                    // dass Zustand nicht mehr bekannt ist, und
                    // keinem Ergebnis -> dann nichts versenden (tritt insbesondere bei den Parametern für
                    // Geschwindigkeitsklassen FG 1 auf, da ein Telegramm auf n Atg abgebildet wird, aber nur eine
                    // Abbildung Daten liefert).
                    if (result != null) {

                        // Jetzt für alle angemeldeten Objekte auf diese Konvertierung einen ResultData
                        // erzeugen und versenden (Austausch des Zielobjekts):
                        SystemObject[ ] objects = _de.getObjectRefsToDe(_config.getType((String) tlsToDav[ OT ]));
                        ResultData[ ]   results = new ResultData[ objects.length ];

                        for (int i = 0; i < objects.length; ++i) {
                            results[ i ] = new ResultData(objects[ i ], result.getDataDescription(), result.getDataTime(), result.getData(), result.isDelayedData());
                        }

                        // Datensätze protokolieren
                        TlsApplication._protTlsReceived.getProtocoller().update(results);
                        debug.fine("Sende Daten an DaV: " + result.getDataDescription().toString());
                        sendData(results);
                        debug.finest("De-Block wurde von Umsetzungsmodul " + this.getClass() + " an DaV versandt!" + Debug.NEWLINE + "Verarbeiteter DeBlock: " + Debug.NEWLINE + deBlock.toString());
                    } else {
                        if (isPassivierungUmsetzungsMethoden()) {

                            // Eventuell sollte wegen Passivierung von Umsetzungsmethoden an dieser DE De Block nicht versendet werden...
                            debug.fine("De-Block wurde ggf. wegen eingeschaltetet Passivierung von Umsetzungsmethoden nicht verarbeitet!" + Debug.NEWLINE + "Umsetzungsmodul-Methode: " + this.getClass() + "-" + _conversionModulName + Debug.NEWLINE + "Nicht ausgebener DeBlock: " + Debug.NEWLINE + deBlock.toString());
                        } else {

                            // Probleme bei der Bearbeitung aufgetreten, nicht verabeiteten DeBlock ausgeben aber weitermachen...
                            debug.warning("De-Block konnte nicht umgesetzt werden (Probleme während der Umsetzung im Umsetzungsmodul)!" + Debug.NEWLINE + "Umsetzungsmodul-Methode: " + this.getClass() + "-" + _conversionModulName + Debug.NEWLINE + "Nicht ausgebener DeBlock: " + Debug.NEWLINE + deBlock.toString());
                        }
                    }
                }
            }

            if (!isConverted) {

                // Wenn kein passender Eintrag für Fg/Typ/Id gefunden wird, nicht verabeiteten
                // DeBlock ausgeben aber weitermachen...
                debug.warning("De-Block wurde im Umsetzungsmodul " + this.getClass() + " nicht verarbeitet, da keine passende Fg/Typ/Id-Kombination gefunden!" + Debug.NEWLINE + "Nicht ausgebener DeBlock: " + Debug.NEWLINE + deBlock.toString());
            }
        }
        catch (NoSuchMethodException e) {
            String signatur = "";

            for (Class _conversionMethodArgument : _conversionMethodArguments) {
                signatur += _conversionMethodArgument + ",";
            }

            debug.warning("De-Block wurde nicht verarbeitet!" + Debug.NEWLINE + "Methodenaufruf [" + _conversionMethodName + "] mit der Signatur (" + signatur + ") " + Debug.NEWLINE + "im UmsetzungsModul [" + _conversionModulName + "] existiert nicht!" + Debug.NEWLINE + "Nicht ausgebener DeBlock: " + Debug.NEWLINE + deBlock.toString() + Debug.NEWLINE + "---------------------------------------------------------------");
        }  // ...bei allen anderen Fehlern Programm terminieren.
        catch (IllegalAccessException e) {
            debug.error("Zugriff auf UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
        }
        catch (InvocationTargetException e) {
            String signatur = "";

            for (Class _conversionMethodArgument : _conversionMethodArguments) {
                signatur += _conversionMethodArgument + ",";
            }

            debug.error("processDeBlock: Methodenaufruf [" + _conversionMethodName + "] am UmsetzungsModul [" + _conversionModulName + "] mit der Signatur (" + signatur + ") fehlgeschlagen!", e);
        }
    }

    /**
     * Sendet einen Ergebnisdatensatz zum Datenverteiler. Die Daten müssen vorher zum Senden angemeldet worden sein.
     *
     * @param result Der zu sendende Ergebnisdatensatz.
     */
    public void sendData(ResultData result) {
        try {
            _daf.sendData(result);
        }
        catch (DataNotSubscribedException e) {
            debug.error("Sendeversuch, obwohl nicht alle Datensätze zum Senden angemeldet sind.", e);
        }
        catch (SendSubscriptionNotConfirmed e) {
            debug.error("Sendeversuch als Sender, obwohl keine Sendesteuerung vorliegt.", e);
        }
    }

    /**
     * Sendet mehrere Ergebnisdatensätze zum Datenverteiler. Die Daten müssen vorher zum Senden angemeldet worden sein.
     *
     * @param results Die zu sendenden Ergebnisdatensätze.
     */
    public void sendData(ResultData[ ] results) {
        try {
            _daf.sendData(results);
        }
        catch (DataNotSubscribedException e) {
            debug.error("Sendeversuch, obwohl nicht alle Datensätze zum Senden angemeldet sind.", e);
        }
        catch (SendSubscriptionNotConfirmed e) {
            debug.error("Sendeversuch als Sender, obwohl keine Sendesteuerung vorliegt.", e);
        }
    }

    /**
     * Methode zum versenden des jeweils letzten Standes eines bereits versandten De-Blocks für jede
     * Knotennummer-DeNummer-Fg-Typ-Id Kombination. Uber diese Methode kann dann z.B. die aktuelle Einstellung für die
     * Kanalsteuerung nach einer De-Gutmeldung verschickt werden. Die Methode greift dabei über den beim speichern
     * verwendeten Hashkey des DeBlocks {@link DeBlock#getLongKey} auf den hier gespeicherten DeBlock zu. Alle
     * versandten De-Blöcke werden dazu nach dem Versand (tatsächlich wenn sie in der Queue zum versenden stehen, ob sie
     * wirklich versandt wurden ist nicht relevant) mit dem {@link DeBlock#getLongKey} gespeichert.
     *
     * @param deBlockKey Der beim Versand verwendete Schlüssel {@link DeBlock#getKey} des De-Blocks
     *                   (Knotennummer-DeNummer-Fg Kombination).
     * @param type       Typ des zu sendenden Telegramms
     * @param id         ID des zu sendenden Telegramms
     */
    public void sendInitialDeBlock(String deBlockKey, int type, int id) {
        deBlockKey += "-" + type + "-" + id;

        DeBlock deBlock = _lastSendDeBlock.get(deBlockKey);

        if (deBlock != null) {
            TlsApplicationLayer._deBlockQueue.push(new DeBlockToSend(_de, deBlock));
            debug.finer("De-Block [Osi7Adresse-DeKanalNummer-Fg-Typ-Id] auf Anfrage versand:" + deBlock.getLongKey());
        } else {

            // Probleme bei der Bearbeiung aufgetreten, Warnung ausgeben aber weitermachen...
            debug.warning("Konnte De-Block für [Osi7Adresse-DeKanalNummer-Fg-Typ-Id] nicht versenden: " + deBlockKey);
        }
    }

    /**
     * Methode zum Versand von Betriebsmeldungen aus einer Umsetzungsmethode heraus. Die Betriebsmeldung wird dabei nur
     * dann versandt, wenn der Parameter zum Versand der Betriebsmeldung für diese DE auf "Ja" eingestellt ist.
     *
     * @param id               ID der Meldung. Dieses Attribut kann von der Applikation gesetzt werden, um einen Bezug
     *                         zu einer vorherigen Meldung herzustellen.
     * @param type             der MeldungsTyp
     * @param messageTypeAddOn der MeldungsTypZusatz
     * @param grade            die MeldungsKlasse
     * @param referenceObject  Referenz auf ein beliebiges Konfigurationsobjekt, auf das sich die Meldung bezieht.
     * @param state            Gibt den Zustand einer Meldung an.
     * @param message          Text der Meldung
     */
    public void sendMessage(String id, MessageType type, String messageTypeAddOn, MessageGrade grade, SystemObject referenceObject, MessageState state, String message) {
        if (_isBetriebsMeldungenErzeugen) {
            _messageSender.sendMessage(id, type, messageTypeAddOn, grade, referenceObject, state, message);
        }
    }

    /**
     * ToDo: Beschreibung der Methode
     *
     *
     * @param tlsToDav TlsToDav
     * @param davToTls DavToTls
     */
    protected void subscribeAllData(Object[ ][ ] tlsToDav, Object[ ][ ] davToTls) {
        DataDescription dataDescription = null;

        // Doppelanmeldung innerhalb eines Umsetzungmoduls erlaubt, da mehrere De-Blöcke auf
        // ein Atg/asp abgebildet werden
        // Fehler nur, wenn zwei Anmeldung auf unterschiedlichen Ebenen im Umsetzungsmodul erfolgen, da
        // dann i.d. R auch zwei Funktionen implementiert sind, die sich (ungewollt ?) überschreiben.
        Map subcribeData = new Hashtable();

        // Zuerst auf interne Struktur umkopieren...
        for (int i = 0; i < tlsToDav.length; ++i) {

            // In Verarbeitungsliste eintragen
            _tlsToDavList.add(tlsToDav[ i ]);
        }

        for (int i = 0; i < davToTls.length; ++i) {

            // Vermeide java.util.ConcurrentModificationException bei gleichzeitigem
            // Zugriff durch die 'update'-Methode
            synchronized (_davToTlsList) {

                // In Verarbeitungsliste auch dann eintragen, wenn DataDescription doppelt (Fg/Typ/id ist anders!)
                _davToTlsList.add(davToTls[ i ]);
            }
        }

        // .. dann die Anmeldungen zum Senden und Empfangen.
        try {

            boolean sendAsSender = false;  // Sendeanmeldung als Sender oder Quelle (default)

            subcribeData.clear();          // Liste für zur Prüfung auf Doppelanmeldungen leeren.

            for (int i = 0; i < tlsToDav.length; ++i) {
                AttributeGroup attributeGroup = _config.getAttributeGroup((String) tlsToDav[ i ][ ATG ]);
                Aspect         aspect         = _config.getAspect((String) tlsToDav[ i ][ ASP ]);

                // Testen, ob ATG und/oder ASP ermittelbar sind, ggf. nicht ermittelbar, weil benötigter KB nicht vorhanden.
                // Dann Fehlermeldung ausgeben und Anmeldung für diesen Eintrag überspringen.
                if(attributeGroup == null || aspect == null) {
                    debug.warning(String.format("Sendeanmeldung (Tls2Dav) nicht möglich, da ATG [%s] und/oder ASP [%s] nicht ermittelbar.", (String) tlsToDav[ i ][ ATG ], (String) tlsToDav[ i ][ ASP ]));
                    continue;
                }

                dataDescription = new DataDescription(attributeGroup, aspect);

                if (subcribeData.get(attributeGroup + "-" + aspect + "-" + tlsToDav[ i ][ OT ]) == null) {
                    debug.finer("Anmeldung zum Senden an den DaV:" + "(Objekt: " + _de.getDe().getNameOrPidOrId() + ") (ATG: " + tlsToDav[ i ][ ATG ] + ") (ASP: " + tlsToDav[ i ][ ASP ] + ") (OT: " + tlsToDav[ i ][ OT ] + ")");

                    // Liste der anzumeldenden Objekte ermittlen
                    SystemObject[ ] objects = _de.getObjectRefsToDe(_config.getType((String) tlsToDav[ i ][ OT ]));

                    // ... und zum Senden anmelden.
                    if (objects.length > 0) {

                        // Testen, ob Aspekt nicht als Quelle sondern als Sender angemeldet werden muss.
                        // ist der Fall bei "konfigurierenden Parametern" der TLS,
                        // Aspekt = "asp.tlsVorgabe"
                        // Aspekt = "asp.tlsAbruf"
                        if ((((String) tlsToDav[ i ][ ASP ]).equals("asp.tlsVorgabe"))
                            || (((String) tlsToDav[ i ][ ASP ]).equals("asp.tlsAbruf"))) {
                            sendAsSender = true;
                        } else {
                            sendAsSender = false;
                        }

                        debug.finer("Anmeldung zum Senden von Daten beim DaV als " + (sendAsSender ? "Sender" : "Quelle") + " :" + "(Objekt: " + _de.getDe().getNameOrPidOrId() + ") (ATG: " + tlsToDav[ i ][ ATG ] + ") (ASP: " + tlsToDav[ i ][ ASP ] + ") (OT: " + tlsToDav[ i ][ OT ] + ")");

                        // ... und zum Senden anmelden.
                        _daf.subscribeSender(this, objects, dataDescription, new SenderRole(!sendAsSender));  // Anmeldung als Sender oder Quelle

                        // In Liste zur Verhinderung von Doppelanmeldungen innerhalb eines Umsetzungsmoduls eintragen
                        subcribeData.put((attributeGroup + "-" + aspect + "-" + tlsToDav[ i ][ OT ]), dataDescription);
                    }
                } else {

                    // COM:debug.finest("Anmeldung in Umsetzungsmodul nicht erfolgt, da bereits angemeldet: " + dataDescription);
                }
            }

            boolean receiveAsDrain = false;  // Empfangsanmeldung als Senke oder Empfänger (default)

            subcribeData.clear();            // Liste für zur Prüfung auf Doppelanmeldungen leeren.

            for (int i = 0; i < davToTls.length; ++i) {
                AttributeGroup attributeGroup = _config.getAttributeGroup((String) davToTls[ i ][ ATG ]);
                Aspect         aspect         = _config.getAspect((String) davToTls[ i ][ ASP ]);

                // Testen, ob ATG und/oder ASP ermittelbar sind, ggf. nicht ermittelbar, weil benötigter KB nicht vorhanden.
                // Dann Fehlermeldung ausgeben und Anmeldung für diesen Eintrag überspringen.
                if(attributeGroup == null || aspect == null) {
                    debug.warning(String.format("Empfangsanmeldung (Dav2Tls) nicht möglich, da ATG [%s] und/oder ASP [%s] nicht ermittelbar.", (String) davToTls[ i ][ ATG ], (String) davToTls[ i ][ ASP ]));
                    continue;
                }

                dataDescription = new DataDescription(attributeGroup, aspect);

                if (subcribeData.get(attributeGroup + "-" + aspect + "-" + davToTls[ i ][ OT ]) == null) {

                    // Liste der anzumeldenden Objekte ermittlen
                    SystemObject[ ] objects = _de.getObjectRefsToDe(_config.getType((String) davToTls[ i ][ OT ]));

                    if (objects.length > 0) {

                        // Testen, ob Aspekt nicht als Empfänger sondern als Senke angemeldet werden muss.
                        // ist der Fall bei "konfigurierenden Parametern" der TLS,
                        // Aspekt = "asp.tlsVorgabe"
                        // Aspekt = "asp.tlsAbruf"
                        if ((davToTls[ i ][ ASP ].equals("asp.tlsVorgabe"))
                            || (davToTls[ i ][ ASP ].equals("asp.tlsAbruf"))) {
                            receiveAsDrain = true;
                        } else {
                            receiveAsDrain = false;
                        }

                        debug.finer("Anmeldung zum Empfang von Daten beim DaV als " + (receiveAsDrain ? "SENKE" : "Empfänger") + " :" + "(Objekt: " + _de.getDe().getNameOrPidOrId() + ") (ATG: " + davToTls[ i ][ ATG ] + ") (ASP: " + davToTls[ i ][ ASP ] + ") (OT: " + davToTls[ i ][ OT ] + ")");

                        // ... und zum empfangen anmelden.
                        _daf.subscribeReceiver(this, objects, dataDescription, new ReceiveOptions(false),  // keine Deltas, keine nachgelieferten Daten
                                               new ReceiverRole(receiveAsDrain));                          // empfangen als Senke oder Empfänger

                        // In Liste zur Verhinderung von Doppelanmeldungen innerhalb eines Umsetzungsmoduls eintragen
                        subcribeData.put((attributeGroup + "-" + aspect + "-" + davToTls[ i ][ OT ]), dataDescription);
                    }
                } else {

                    // COM:debug.finest("Anmeldung in Umsetzungsmodul nicht erfolgt, da bereits angemeldet: " + dataDescription);
                }
            }

            // Bei Clusterkanälen der FG 4 zusätzlich Attributgruppe zur Pufferabfrage als SENDER anmelden
            // Wird bei Folgennummerfehlern benötigt.
            if (!_isFg4AngemeldetAlsCluster
                && (_de.getFg() == 4)
                && _de.isCluster()) {
                AttributeGroup attributeGroup = _config.getAttributeGroup("atg.tlsWzgAbrufPufferInhalt");
                Aspect         aspect         = _config.getAspect("asp.tlsAbruf");

                dataDescription = new DataDescription(attributeGroup, aspect);
                _daf.subscribeSender(this, _de.getDe(), dataDescription, new SenderRole(false));           // Anmeldung als Sender
                _isFg4AngemeldetAlsCluster = true;                                                         // Anmeldung darf nur einmal erfolgen
            }

            // Bei Sammelkanälen der FG 4 zusätzlich Attributgruppe zur Abfrage von De-Blöcken als SENDER anmelden.
            // Wird bei Folgennummerfehlern benötigt, um aktuellen Zustand anzuforden.
            if (!_isFg4AngemeldetAlsSammelKanal
                && (_de.getFg() == 4)) {
                AttributeGroup attributeGroup = _config.getAttributeGroup("atg.tlsGloDeBlockAbruf");
                Aspect         aspect         = _config.getAspect("asp.tlsAbruf");

                dataDescription = new DataDescription(attributeGroup, aspect);
                _daf.subscribeSender(this, _de.getDe(), dataDescription, new SenderRole(false));           // Anmeldung als Sender
                _isFg4AngemeldetAlsSammelKanal = true;                                                     // Anmeldung darf nur einmal erfolgen
            }

            // Anmeldung des Parameters zur Verwaltung der Passivierung von Umsetzungsmethoden
            // Falls Anmeldung fehlschlägt (z. B. nicht aktueller kb.tmKExTlsGlobal, V21 oder höher)
            try {
                if (!_isAngemeldetPassivierungUmsetzungsMethoden) {
                    AttributeGroup attributeGroup = _config.getAttributeGroup("atg.tlsPassivierungÜbernahmeLogischesModell");
                    Aspect         aspect         = _config.getAspect("asp.parameterSoll");

                    dataDescription = new DataDescription(attributeGroup, aspect);
                    _daf.subscribeReceiver(this, _de.getDe(), dataDescription, new ReceiveOptions(false),  // keine Deltas, keine nachgelieferten Daten
                                           new ReceiverRole(false));                                       // Anmeldung als Empfänger
                    _isAngemeldetPassivierungUmsetzungsMethoden = true;                                    // Anmeldung darf nur einmal erfolgen
                }
            }
            catch (Exception e) {
                debug.warning("Anmeldung auf atg.tlsPassivierungÜbernahmeLogischesModell am DE fehlgeschlagen. Benötigt kb.tmKExTlsGlobal ab V21.");
            }

            // Anmeldung des Parameters zur Verwaltung des Versands von Betriebsmeldungen in Umsetzungsmethoden
            // Falls Anmeldung fehlschlägt (z. B. nicht aktueller kb.tmKExTlsGlobal, V21 oder höher)
            try {
                if (!_isAngemeldetBetriebsMeldungenErzeugen) {
                    AttributeGroup attributeGroup = _config.getAttributeGroup("atg.tlsBetriebsMeldungErzeugen");
                    Aspect         aspect         = _config.getAspect("asp.parameterSoll");

                    dataDescription = new DataDescription(attributeGroup, aspect);
                    _daf.subscribeReceiver(this, _de.getDe(), dataDescription, new ReceiveOptions(false),  // keine Deltas, keine nachgelieferten Daten
                                           new ReceiverRole(false));                                       // Anmeldung als Empfänger
                    _isAngemeldetBetriebsMeldungenErzeugen = true;                                         // Anmeldung darf nur einmal erfolgen
                }
            }
            catch (Exception e) {
                debug.warning("Anmeldung auf atg.tlsBetriebsMeldungErzeugen am DE fehlgeschlagen. Benötigt kb.tmKExTlsGlobal ab V21.");
            }
        }
        catch (OneSubscriptionPerSendData e) {
            debug.error("Daten wurden bereits als Quelle angemeldet: " + dataDescription, e);
        }

        debug.finest("UmwandlungsModul " + this.getClass().toString() + " wurde erfolgreich geladen " + "und instanziert für DE: " + Debug.NEWLINE + _de.toString());
    }

    // ------------------------------------------------------------------------
    // INTERFACE METHODEN
    // ------------------------------------------------------------------------
    // --------------------- Interface ClientReceiverInterface ---------------------

    /**
     * Aktualisierungsmethode, die nach Empfang eines angemeldeten Datensatzes von den
     * Datenverteiler-Applikationsfunktionen aufgerufen wird. Analysiert die Ergebnisdatensätze und ruft die
     * entsprechende Methode zur weiteren Verarbeitung (Zusammenstellung eines DE-Blocks) auf.
     *
     * @param results Feld mit den empfangenen Ergebnisdatensätzen.
     */
    public void update(ResultData[ ] results) {
        ResultData      result = null;
        DataDescription dataDescription;

        // Datensätze protokolieren
        TlsApplication._protTlsSend.getProtocoller().update(results);

        try {

            // Datensatz analysieren und geeignete Methode aufrufen....
            for (int i = 0; i < results.length; i++) {
                result = results[ i ];

                if (result.hasData()) {
                    debug.finer(results.length + " Ergebnisdatensätze erhalten.");
                    debug.finest("Daten empfangen von: " + result.getObject().getNameOrPidOrId() + Debug.NEWLINE + "Datenbeschreibung  : " + result.getDataDescription() + Debug.NEWLINE + "Daten              : " + result.getData().toString());

                    // Jetzt alle für diese DE angemeldeten Umsetzungsmethoden auswerten
                    boolean isConverted = false;  // wurde Datensatz mindestens einmal konvertiert/ausgewertet ?

                    // Zuerst prüfen, ob empfangener Datensatz Paramtersatz für
                    // - Betriebsmeldungen ausgeben (ja/nein)
                    // - Passivierung Umsetzungsmethoeden (ja/nein)
                    // ist, erst danach die angemeldeten Umsetzungsmethoden auswerten
                    // BetriebsMeldungenErzeugen (ja/nein)
                    dataDescription = new DataDescription(_config.getAttributeGroup("atg.tlsBetriebsMeldungErzeugen"), _config.getAspect("asp.parameterSoll"));

                    if ((dataDescription.getAttributeGroup().equals(result.getDataDescription().getAttributeGroup()))
                        && (dataDescription.getAspect().equals(result.getDataDescription().getAspect()))) {
                        _isBetriebsMeldungenErzeugen = result.getData().getItem("BetriebsMeldungErzeugen").asTextValue().getText().equals("Ja");
                        isConverted                  = true;
                    }

                    // Umsetzungsmethoden passivieren (ja/nein)
                    dataDescription = new DataDescription(_config.getAttributeGroup("atg.tlsPassivierungÜbernahmeLogischesModell"), _config.getAspect("asp.parameterSoll"));

                    if ((dataDescription.getAttributeGroup().equals(result.getDataDescription().getAttributeGroup()))
                        && (dataDescription.getAspect().equals(result.getDataDescription().getAspect()))) {
                        _isPassivierungUmsetzungsMethoden = result.getData().getItem("PassivierungÜbernahmeLogischesModell").asTextValue().getText().equals("Ja");
                        isConverted                       = true;
                    }

                    // Vermeide java.util.ConcurrentModificationException bei gleichzeitigem
                    // Zugriff durch die 'subscribeAllData'-Methode
                    synchronized (_davToTlsList) {
                        Iterator<Object[ ]> iterator = _davToTlsList.iterator();

                        while (iterator.hasNext()) {
                            Object[ ]      davToTls       = iterator.next();
                            AttributeGroup attributeGroup = _config.getAttributeGroup((String) davToTls[ ATG ]);
                            Aspect         aspect         = _config.getAspect((String) davToTls[ ASP ]);

                            dataDescription = new DataDescription(attributeGroup, aspect);

                            if ((dataDescription.getAttributeGroup().equals(result.getDataDescription().getAttributeGroup()))
                                && (dataDescription.getAspect().equals(result.getDataDescription().getAspect()))) {
                                isConverted = true;  // mindestens eine Umsetzungsmethode gefunden

                                // Methodenname bilden (z.B.) Fg001Typ029Id002AtgTlsGloKanalSteuerungAspTlsAntwort(...);
                                _conversionMethodName = buildMethodName(davToTls);
                                _conversionModulName  = this.toString();
                                debug.finest("Ermitteter Methodenname für Umsetzung von Objekt " + result.getObject().getNameOrPidOrId() + ": " + _conversionMethodName);
                                _conversionMethodArguments = new Class[ ] { ResultData.class };
                                _conversionMethod          = this.getClass().getMethod(_conversionMethodName, _conversionMethodArguments);

                                // es wird erlaubt, dass die Conversionmethode eine <List> von DE-Bloecken liefert
                                // deBlock = (DeBlock) _conversionMethod.invoke(this, new Object[]{result});
                                // neue Verarbeitung
                                Object  conversionResult = _conversionMethod.invoke(this, new Object[ ] { result });
                                List<?> deBlockList;

                                if (conversionResult instanceof List) {
                                    deBlockList = (List<?>) conversionResult;
                                } else {

                                    // Dummy-Liste zur Vermeidung von 'unchecked cast' Warnungen
                                    List<Object> deBlockListDummy = new ArrayList<Object>();

                                    deBlockListDummy.add(conversionResult);
                                    deBlockList = (List<?>) deBlockListDummy;
                                }

                                Iterator<?> deBlockIterator = deBlockList.iterator();

                                while (deBlockIterator.hasNext()) {
                                    Object bo = deBlockIterator.next();

                                    if (!(bo instanceof DeBlock)) {
                                        debug.warning("Conversion-Ergebnis ist nicht vom Typ 'DeBlock'");

                                        continue;
                                    }

                                    DeBlock deBlock = (DeBlock) bo;

                                    // Ende Gieseler, BitCtrl neue Verarbeitung
                                    // Prüfen, ob DE-Block überhaupt versendet werden darf...
                                    if (isAllowedToSend(_de, deBlock)) {
                                        if (deBlock != null) {

                                            // DeBlock und zugeordnetes De-Objekt zum Versenden weiterleiten.
                                            // Zuerst prüfen, ob DE-Block für späteren getriggerten Versand  von mehreren De-Blöcken
                                            // in einem Einzeltelegramm zwischengespeichert werden soll, ansonsten direkt versenden.
                                            if (_collectDeBlocks.isDeBlockToBuffer(deBlock)) {
                                                _de.getEak().bufferDeBlock(new DeBlockToSend(_de, deBlock));
                                            } else {
                                                TlsApplicationLayer._deBlockQueue.push(new DeBlockToSend(_de, deBlock));
                                            }

                                            // De-Block für Sequenzen merken, z.B. bei De-Gut-Meldung Betriebsparameter und Kanal
                                            // Steuerung senden (Schlüssel: Osi7Adresse-De-Fg-Typ-Id).
                                            debug.finest("De-Block [Osi7Adresse-DeKanalNummer-Fg-Typ-Id] für " + "erneuten Versand gespeichert: " + deBlock.getLongKey());
                                            _lastSendDeBlock.put(deBlock.getLongKey(), deBlock);
                                        } else {

                                            // Probleme bei der Bearbeiung aufgetreten, Warnung ausgeben aber weitermachen...
                                            debug.warning("Empfangene Attributgruppe konnte nicht umgesetzt werden: " + dataDescription);
                                        }
                                    }
                                }
                            }
                        }
                    }  // synchronized

                    if (!isConverted) {

                        // Wenn Methode nicht gefunden wird, Warnung ausgeben aber weitermachen...
                        debug.warning("Für empfangene Attributgruppe ist keine Weiterverarbeitung in " + this.getClass() + " vorgesehen: " + result.getDataDescription());
                    }
                }
            }
        }
        catch (NoSuchMethodException e) {
            String signatur = "";

            for (Class _conversionMethodArgument : _conversionMethodArguments) {
                signatur += _conversionMethodArgument + ",";
            }

            debug.warning("Ergebnisdaten wurde nicht verarbeitet!" + Debug.NEWLINE + "Methodenaufruf [" + _conversionMethodName + "] mit der Signatur (" + signatur + ") " + Debug.NEWLINE + "im UmsetzungsModul [" + _conversionModulName + "] existiert nicht!" + Debug.NEWLINE + "Nicht versandte Daten: " + Debug.NEWLINE + "Datenbeschreibung  : " + result.getDataDescription() + Debug.NEWLINE + "Daten              : " + result.getData().toString() + Debug.NEWLINE +
                          "---------------------------------------------------------------");
        }  // ...bei allen anderen Fehlern Programm terminieren.
        catch (IllegalAccessException e) {
            debug.error("Zugriff auf UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
        }
        catch (IllegalArgumentException e) {
            String signatur = "";

            for (Class _conversionMethodArgument : _conversionMethodArguments) {
                signatur += _conversionMethodArgument + ",";
            }

            debug.error("Methodenaufruf [" + _conversionMethodName + "] mit der Signatur (" + signatur + ") im UmsetzungsModul [" + _conversionModulName + "] existiert nicht!", e);
        }
        catch (InvocationTargetException e) {
            String signatur = "";

            for (Class _conversionMethodArgument : _conversionMethodArguments) {
                signatur += _conversionMethodArgument + ",";
            }

            debug.error("update: Methodenaufruf [" + _conversionMethodName + "] am UmsetzungsModul [" + _conversionModulName + "] mit der Signatur (" + signatur + ") fehlgeschlagen!", e);
        }
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 08.02.17
